/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/posix/thread.c: Thread management for POSIX-compatible systems.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/posix/thread.h"

#include <pthread.h>
#include <sched.h>
#include <unistd.h>

#ifdef __linux__  // Any Linux-based platform (including Android).
# include <sys/resource.h>
#endif

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Priority of the initial thread, for adjusting priorities in
 * sys_thread_create(). */
static int base_priority;

/* Local storage key for storing the current thread's SysThread pointer
 * (so we can return it from sys_thread_get_id()). */
static pthread_key_t sys_thread_key;

/* Control object for creating the local storage key. */
static pthread_once_t key_once = PTHREAD_ONCE_INIT;

/* Flag indicating whether the key was successfully created (since we can't
 * return a value from the creation function). */
static uint8_t sys_thread_key_created = 0;

/*-----------------------------------------------------------------------*/

/* Local routine declarations. */

/**
 * posix_thread_init:  Save the current thread's priority, and create the
 * local storage key for storing the current thread's SysThread pointer.
 */
static void posix_thread_init(void);

/**
 * get_raw_priority:  Retrieve the raw system-level priority value for the
 * current thread.  Wraps the difference between Linux and standard POSIX
 * priority handling (excluding the inverted priority sign).
 *
 * [Parameters]
 *     priority_ret: Pointer to variable to receive the current thread's
 *         raw priority value.
 * [Return value]
 *     True on success, false on error.
 */
static int get_raw_priority(int *priority_ret);

/**
 * thread_runner:  Wrapper for threads started with sys_thread_create()
 * which handles setting the "finished" flag when the thread terminates
 * and returning a pointer as pthreads expects.  On Android, this function
 * also attaches the thread to / detaches it from the Dalvik VM.
 *
 * [Parameters]
 *     param: Thread parameter (points to a SysThread structure).
 * [Return value]
 *     Thread function return value.
 */
static void *thread_runner(void *param);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

SysThreadID sys_thread_create(
    int priority, int stack_size, int (*function)(void *), void *param)
{
    #define CHECK(call)  do {                       \
        const int error = (call);                   \
        if (UNLIKELY(error != 0)) {                 \
            DLOG("%s: %s", #call, strerror(error)); \
            return 0;                               \
        }                                           \
    } while (0)

    CHECK(pthread_once(&key_once, posix_thread_init));
    if (UNLIKELY(!sys_thread_key_created)) {
        DLOG("Failed to create TLS key for thread ID");
        return 0;
    }

    int policy;
    struct sched_param sched_param;
    CHECK(pthread_getschedparam(pthread_self(), &policy, &sched_param));

#ifdef __linux__
    /* Linux (including Android) doesn't allow setting thread priorities
     * via pthreads, but it usefully deviates from POSIX in allowing
     * setpriority() to work on single threads, so we take that approach
     * instead. */
    int real_priority = base_priority - priority;
# ifdef SIL_PLATFORM_ANDROID
    const int pri_min = -20;
# else  // SIL_PLATFORM_LINUX
    int pri_min;
    struct rlimit rlim;
    if (UNLIKELY(getrlimit(RLIMIT_NICE, &rlim) != 0)) {
        DLOG("%p(%p): getrlimit(RLIMIT_NICE): %s", function, param,
             strerror(errno));
    }
    if (rlim.rlim_cur == RLIM_INFINITY) {
        pri_min = -20;
    } else {
        pri_min = 20 - bound(rlim.rlim_cur, 1, 40);
    }
    /* We can always start a thread at the same or lower priority (higher
     * nice level) regardless of resource limit settings. */
    int my_priority;
    if (get_raw_priority(&my_priority) && my_priority < pri_min) {
        pri_min = my_priority;
    }
# endif
    const int pri_max = 19;
    if (real_priority < pri_min) {
        DLOG("%p(%p): Requested priority %d (actual %d) too high, using"
             " %d (%d)", function, param, priority, real_priority,
             priority - (pri_min - real_priority), pri_min);
        real_priority = pri_min;
    } else if (real_priority > pri_max) {
        DLOG("%p(%p): Requested priority %d (actual %d) too low, using"
             " %d (%d)", function, param, priority, real_priority,
             priority - (pri_max - real_priority), pri_max);
        real_priority = pri_max;
    }
#else  // !__linux__
    sched_param.sched_priority = base_priority + priority;
    const int pri_min = sched_get_priority_min(policy);
    const int pri_max = sched_get_priority_max(policy);
    if (sched_param.sched_priority < pri_min) {
        DLOG("%p(%p): Requested priority %d (actual %d) too low, using %d"
             " (%d)", function, param, priority, sched_param.sched_priority,
             priority + (pri_min - sched_param.sched_priority), pri_min);
        sched_param.sched_priority = pri_min;
    } else if (sched_param.sched_priority > pri_max) {
        DLOG("%p(%p): Requested priority %d (actual %d) too high, using %d"
             " (%d)", function, param, priority, sched_param.sched_priority,
             priority + (pri_max - sched_param.sched_priority), pri_max);
        sched_param.sched_priority = pri_max;
    }
#endif  // __linux__

    pthread_attr_t attr;
    CHECK(pthread_attr_init(&attr));
    CHECK(pthread_attr_setschedpolicy(&attr, policy));
    CHECK(pthread_attr_setschedparam(&attr, &sched_param));
    if (stack_size > 0) {
        int page_size = sysconf(_SC_PAGESIZE);
        ASSERT(page_size > 0, page_size = 4096);
        stack_size = align_up(stack_size, page_size);
        stack_size = lbound(stack_size, (int)PTHREAD_STACK_MIN);
#ifdef __linux__
        /* Linux doesn't let you actually access the lowest page of the
         * stack (WTF?), and we don't seem to get to use the highest page,
         * so add two extra pages to the requested stack size to ensure we
         * get the usable size we want. */
        stack_size += page_size*2;
#endif
#ifdef SIL_PLATFORM_IOS
        /* Add an extra page to cover stack usage by DLOG() -> NSLog(). */
        stack_size += page_size;
#endif
#ifdef COVERAGE
        /* Add extra space to account for thread-local storage for
         * coverage tracking. */
        stack_size += page_size*4;
#endif
        CHECK(pthread_attr_setstacksize(&attr, stack_size));
    }

    SysThread *thread = mem_alloc(sizeof(*thread), 0, 0);
    if (UNLIKELY(!thread)) {
        DLOG("No memory for thread structure");
        return 0;
    }
    thread->function = function;
    thread->param = param;
#ifdef __linux__
    thread->priority = real_priority;
#endif
    thread->finished = 0;

    const int error = pthread_create(&thread->handle, &attr,
                                     thread_runner, thread);
    pthread_attr_destroy(&attr);
    if (UNLIKELY(error != 0)) {
        DLOG("Failed to create thread for %p(%p): %s", function, param,
             strerror(error));
        mem_free(thread);
        return 0;
    }

    return (SysThreadID)thread;

    #undef CHECK
}

/*-----------------------------------------------------------------------*/

void sys_thread_exit(int exit_code)
{
    posix_thread_runner_cleanup(pthread_getspecific(sys_thread_key));
    pthread_exit((void *)(intptr_t)exit_code);
}

/*-----------------------------------------------------------------------*/

SysThreadID sys_thread_get_id(void)
{
    return (SysThreadID)pthread_getspecific(sys_thread_key);
}

/*-----------------------------------------------------------------------*/

int sys_thread_get_priority(void)
{
    int my_priority;
    if (UNLIKELY(!get_raw_priority(&my_priority))) {
        return 0;
    }
#ifdef __linux__
    return base_priority - my_priority;
#else
    return my_priority - base_priority;
#endif
}

/*-----------------------------------------------------------------------*/

int sys_thread_is_running(SysThreadID thread_)
{
    SysThread *thread = (SysThread *)thread_;
    BARRIER();
    return !thread->finished;
}

/*-----------------------------------------------------------------------*/

int sys_thread_wait(SysThreadID thread_, int *result_ret)
{
    SysThread *thread = (SysThread *)thread_;
#ifdef SIL_PLATFORM_ANDROID
    /* Some older versions of Android fail to detect waiting on self. */
    if (UNLIKELY(thread->handle == pthread_self())) {
        DLOG("pthread_join(%p): %s", &thread->handle, strerror(EDEADLK));
        return 0;
    }
#endif
    void *retval = NULL;
    int error = pthread_join(thread->handle, &retval);
    if (UNLIKELY(error != 0)) {
        DLOG("pthread_join(%p): %s", &thread->handle, strerror(error));
        return 0;
    }
    mem_free(thread);
    /* Convert through intptr_t to silence compiler warnings about
     * converting from a pointer to a differently-sized integer. */
    *result_ret = (int)(intptr_t)retval;
    return 1;
}

/*-----------------------------------------------------------------------*/

void sys_thread_yield()
{
    sched_yield();
}

/*************************************************************************/
/*********************** Exported utility routines ***********************/
/*************************************************************************/

int posix_thread_create_detached(void (*function)(void *), void *param)
{
    PRECOND(function != NULL, return 0);

    #define CHECK(call)  do {                       \
        const int error = (call);                   \
        if (UNLIKELY(error != 0)) {                 \
            DLOG("%s: %s", #call, strerror(error)); \
            return 0;                               \
        }                                           \
    } while (0)

    CHECK(pthread_once(&key_once, posix_thread_init));
    if (UNLIKELY(!sys_thread_key_created)) {
        DLOG("Failed to create TLS key for thread ID");
        return 0;
    }

    pthread_attr_t attr;
    CHECK(pthread_attr_init(&attr));
    CHECK(pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED));

    pthread_t handle;
    const int error = pthread_create(&handle, &attr,
                                     (void *(*)(void *))function, param);
    pthread_attr_destroy(&attr);
    if (LIKELY(error == 0)) {
        return 1;
    } else {
        DLOG("Failed to create thread for %p(%p): %s", function, param,
             strerror(error));
        return 0;
    }

    #undef CHECK
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static void posix_thread_init(void)
{
    errno = 0;
    if (UNLIKELY(!get_raw_priority(&base_priority))) {
        base_priority = 0;
    }

    int error;
    if (UNLIKELY((error = pthread_key_create(&sys_thread_key, NULL)) != 0)) {
        DLOG("pthread_key_create() failed: %s", strerror(error));
        return;
    }
    sys_thread_key_created = 1;
}

/*-----------------------------------------------------------------------*/

static int get_raw_priority(int *priority_ret)
{
    ASSERT(priority_ret != NULL, return 0);

#ifdef __linux__
    errno = 0;
    const int my_priority = getpriority(PRIO_PROCESS, 0);
    if (my_priority == -1 && UNLIKELY(errno != 0)) {
        DLOG("getpriority(PRIO_PROCESS, 0) failed: %s", strerror(errno));
        return 0;
    }
    *priority_ret = my_priority;
#else
    int policy;
    struct sched_param sched_param;
    int error = pthread_getschedparam(pthread_self(), &policy, &sched_param);
    if (UNLIKELY(error != 0)) {
        DLOG("pthread_getschedparam(self) failed: %s", strerror(error));
        return 0;
    }
    *priority_ret = sched_param.sched_priority;
#endif
    return 1;
}

/*-----------------------------------------------------------------------*/

static void *thread_runner(void *param)
{
    SysThread *thread = (SysThread *)param;
    ASSERT(thread != NULL, return NULL);

    int error;
    if (UNLIKELY((error = pthread_setspecific(sys_thread_key, thread)) != 0)) {
        DLOG("Failed to store thread ID");
        return NULL;
    }

    posix_thread_runner_init(thread);
    const int retval = (*thread->function)(thread->param);
    posix_thread_runner_cleanup(thread);

    /* Make sure all actions from the thread function happen before the
     * thread->finished store is seen. */
    BARRIER();
    thread->finished = 1;

    return (void *)(intptr_t)retval;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
