/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/windows/internal.h: Declarations used internally by
 * Windows-specific code.
 */

#ifndef SIL_SRC_SYSDEP_WINDOWS_INTERNAL_H
#define SIL_SRC_SYSDEP_WINDOWS_INTERNAL_H

/*************************************************************************/
/*************************************************************************/

/* Windows version constants, for use with WINVER and the windows_version*
 * functions. */

#define WINDOWS_VERSION_2000   0x0500
#define WINDOWS_VERSION_XP     0x0501
#define WINDOWS_VERSION_VISTA  0x0600
#define WINDOWS_VERSION_7      0x0601
#define WINDOWS_VERSION_8      0x0602
#define WINDOWS_VERSION_8_1    0x0603

/*-----------------------------------------------------------------------*/

/* System headers.  Note that we need to do a bit of magic with #defines
 * because windef.h tries to define "near" and "far" to nothing. */

#undef near
#undef far

#include <windows.h>
#include <windowsx.h>
#include <shlobj.h>

#if defined(__MINGW32__) && !defined(__MINGW64__)  // 32-bit MinGW has these in a wrong/obsolete place?
# include <ddk/hidusage.h>
# include <ddk/hidpi.h>
# include <ddk/hidsdi.h>
#else
# include <hidusage.h>
# include <hidpi.h>
# include <hidsdi.h>
#endif

/* Fix Windows stomping on our macros. */
#undef ASSERT
#ifdef DEBUG
# define ASSERT(condition,...)                                          \
    if (UNLIKELY(!(condition))) {                                       \
        do_DLOG(__FILE__, __LINE__, __FUNCTION__,                       \
                "\n\n*** ALERT *** ASSERTION FAILED:\n%s\n\n",          \
                #condition);                                            \
        if (0) {__VA_ARGS__;}                                           \
        abort();                                                        \
    } else (void)0
#else
# define ASSERT(condition,...)          \
    if (UNLIKELY(!(condition))) {       \
        __VA_ARGS__;                    \
    } else (void)0
#endif

#ifdef __MINGW32__
# define ENDSESSION_CRITICAL  0x40000000  // Missing from MinGW headers.
#endif

#if WINVER < WINDOWS_VERSION_VISTA
# define WM_MOUSEHWHEEL  0x20E
# ifndef __MINGW64__  // Not suppressed in 64-bit MinGW headers for pre-Vista.
#  define GIDC_ARRIVAL  1
#  define GIDC_REMOVAL  2
#  define RIDEV_DEVNOTIFY  0x2000
#  define RIDEV_EXINPUTSINK  0x1000
#  define WM_INPUT_DEVICE_CHANGE  0xFE
typedef struct _RTL_CONDITION_VARIABLE {
    PVOID Ptr;
} CONDITION_VARIABLE, *PCONDITION_VARIABLE;
extern VOID WINAPI InitializeConditionVariable(
    PCONDITION_VARIABLE ConditionVariable);
extern BOOL WINAPI SleepConditionVariableCS(
    PCONDITION_VARIABLE ConditionVariable, PCRITICAL_SECTION CriticalSection,
    DWORD dwMilliseconds);
extern VOID WINAPI WakeAllConditionVariable(
    PCONDITION_VARIABLE ConditionVariable);
extern VOID WINAPI WakeConditionVariable(
    PCONDITION_VARIABLE ConditionVariable);
# endif
#endif

#if WINVER < WINDOWS_VERSION_8
# define SM_DIGITIZER  94
# define NID_INTEGRATED_TOUCH  0x01
# define WM_POINTERUPDATE  0x0245
# define WM_POINTERDOWN  0x0246
# define WM_POINTERUP  0x0247
# define WM_POINTERCAPTURECHANGED  0x024C
# define GET_POINTERID_WPARAM(wParam)  LOWORD((wParam))
typedef enum {
    PT_POINTER = 1,
    PT_TOUCH = 2,
    PT_PEN = 3,
    PT_MOUSE = 4,
    PT_TOUCHPAD = 5,
} POINTER_INPUT_TYPE;
extern BOOL WINAPI GetPointerType(UINT32 pointerId,
                                  POINTER_INPUT_TYPE *pointerType);
#endif

#if WINVER >= WINDOWS_VERSION_8_1
# include <ShellScalingApi.h>
#else
typedef enum {
    PROCESS_DPI_UNAWARE = 0,
    PROCESS_SYSTEM_DPI_AWARE = 1,
    PROCESS_PER_MONITOR_DPI_AWARE = 2,
} PROCESS_DPI_AWARENESS;
#endif

#include "src/sysdep/windows/utf8-wrappers.h"

#undef near
#undef far
#define near near_
#define far far_

/* Also get rid of these.  Will we ever be free from win16 crap? */
#undef NEAR
#undef FAR

/*-----------------------------------------------------------------------*/

/* Library-internal declarations. */


/******** condvar.c ********/

/**
 * TEST_windows_condvar_disable_native:  When set to true (nonzero), the
 * sys_condvar_*() functions will use emulated condition variables even if
 * the system supports native condition variables.
 */
extern uint8_t TEST_windows_condvar_disable_native;


/******** graphics.c ********/

/*
 * Wrappers for wgl*() functions (see definitions in graphics.c for
 * rationale).  We have to redeclare these to override the dllimport
 * declarations in <wingdi.h>.
 */

/* Suppress dllimport warnings, since we deliberately redeclare these. */
#ifdef __GNUC__
# pragma GCC diagnostic push
# pragma GCC diagnostic ignored "-Wattributes"
#endif

extern HGLRC WINAPI wglCreateContext(HDC dc);
extern BOOL WINAPI wglDeleteContext(HGLRC context);
extern HGLRC WINAPI wglGetCurrentContext(void);
extern PROC WINAPI wglGetProcAddress(const char *name);
extern BOOL WINAPI wglMakeCurrent(HDC dc, HGLRC context);

#ifdef __GNUC__
# pragma GCC diagnostic pop
#endif

/**
 * windows_close_window:  Close the currently open window, if any.
 * Intended primarily for use by test routines.
 */
extern void windows_close_window(void);

/**
 * windows_reset_video_mode:  If the video mode of the display device has
 * been changed, reset it to the original mode.
 */
extern void windows_reset_video_mode(void);

/**
 * windows_set_mouse_grab:  Set whether to confine the mouse to the window
 * while a window is open.
 *
 * [Parameters]
 *     grab: True to grab the mouse, false to not grab.
 */
extern void windows_set_mouse_grab(int grab);

/**
 * windows_show_mouse_pointer:  Override or restore the shown/hidden state
 * of the mouse pointer.
 *
 * [Parameters]
 *     override: 1 to show the mouse pointer, 0 to hide the pointer, -1 to
 *         restore the state selected by sys_graphics_show_mouse_pointer().
 */
extern void windows_show_mouse_pointer(int override);

/**
 * windows_wgl_context:  Return the GL context for the current window.
 *
 * [Return value]
 *     GL context for the current window, or NULL if no window is open.
 */
extern HGLRC windows_wgl_context(void);

/**
 * windows_window:  Return the handle for the current window.
 *
 * [Return value]
 *     Handle for the current window, or NULL if no window is open.
 */
extern HWND windows_window(void);

/**
 * windows_window_title:  Return the window title currently set via
 * graphics_set_window_title().
 *
 * [Return value]
 *     Window title, or NULL if no title has been set.
 */
extern const char *windows_window_title(void);


/******** input.c ********/

/**
 * windows_handle_input_message:  Handle window messages relating to input
 * events.
 *
 * [Parameters]
 *     hwnd: Window on which the message was received (always the currently
 *         open window).
 *     uMsg: Message type.
 *     wParam, lParam: Message parameters.
 * [Return value]
 *     True if the message was consumed, false if not.
 */
extern int windows_handle_input_message(HWND hwnd, UINT uMsg,
                                        WPARAM wParam, LPARAM lParam);


/******** main.c ********/

/**
 * windows_executable_dir:  Return the pathname of the directory containing
 * the executable file used to start the program, or "." if the directory
 * is unknown.  The path is returned in Windows format, using backslashes
 * rather than slashes to separate path components.
 */
extern const char *windows_executable_dir(void);


/******** misc.c ********/

/**
 * windows_set_error:  Record the given SYSERR_* error code as the error to
 * be returned from sys_last_error().  If code == 0, the code is instead
 * set based on the Windows error code specified by windows_code; if
 * windows_code is also zero, the current Windows error code (as returned
 * by GetLastError()) is used.
 *
 * [Parameters]
 *     code: SYSERR_* error code, or zero to use a Windows error code.
 *     windows_code: Windows error code, or zero to use the system error code.
 */
extern void windows_set_error(int code, DWORD windows_code);


/******** util.c ********/

/**
 * strdup_16to8:  Translate the given UTF-16 string into UTF-8 and return
 * it in a buffer allocated with mem_alloc().  Any invalid surrogates (a
 * lead surrogate not followed by a trail surrogate or vice versa) are
 * replaced with U+FFFD.
 *
 * [Parameters]
 *     s16: UTF-16 string to convert (must be non-NULL).
 * [Return value]
 *     Converted string, or NULL if out of memory.
 */
extern char *strdup_16to8(const uint16_t *s16);

/**
 * convert_16to8:  Translate the given UTF-16 string into UTF-8 in the same
 * manner as strdup_16to8(), and store it in the given buffer.
 *
 * This function does not check for buffer overflow; the caller is
 * responsible for ensuring that the buffer is large enough to hold the
 * result of conversion.
 *
 * [Parameters]
 *     s16: UTF-16 string to convert (must be non-NULL).
 *     buffer: Buffer to hold output string (must have at least 3 bytes of
 *         space for every 16-byte element in the source string, plus one
 *         byte for the trailing null).
 */
extern void convert_16to8(const uint16_t *s16, char *buffer);

/**
 * strdup_8to16:  Translate the given UTF-8 string into UTF-16 and return
 * it in a buffer allocated with mem_alloc().  Invalid UTF-8 sequences are
 * ignored; codepoints which cannot be encoded in UTF-16 (U+D800 through
 * U+DFFF, U+110000 and above) are replaced with U+FFFD.
 *
 * [Parameters]
 *     s8: UTF-8 string to convert (must be non-NULL).
 * [Return value]
 *     Converted string, or NULL if out of memory.
 */
extern uint16_t *strdup_8to16(const char *s8);

/**
 * timeout_to_ms:  Convert the given timeout to a value suitable for
 * passing to WaitForSingleObject() or similar wait functions.  Helper
 * function for sys_condvar_wait(), sys_mutex_lock(), and sys_semaphore_wait().
 *
 * [Parameters]
 *     timeout: Timeout, in seconds, or a negative value for no timeout.
 * [Return value]
 *     Value to pass as the dwMilliseconds parameter to WaitForSingleObject().
 */
extern DWORD timeout_to_ms(float timeout);

/**
 * windows_getenv:  Retrieve an environment variable's value.  The returned
 * buffer should be freed with mem_free() when no longer needed.
 *
 * [Parameters]
 *     name: Name of environment variable to retrieve.
 * [Return value]
 *     Value of environment variable, or NULL if the environment variable
 *     is not found (or memory allocation failed).
 */
extern char *windows_getenv(const char *name);

/**
 * windows_strerror:  Return a text string describing the error associated
 * with the given Windows error code.  The string is stored in a static
 * buffer which will be overwritten on each call to the function.
 *
 * [Parameters]
 *     code: Windows error code.
 * [Return value]
 *     Error description (including the error code at the beginning).
 */
extern const char *windows_strerror(DWORD code);

/**
 * windows_version:  Return the version of Windows under which the program
 * is running, as a WINVER-style 16-bit value (see the WINDOWS_VERSION_*
 * constants).
 *
 * Due to limitations of the Windows API, this function currently will not
 * return any version number greater than WINDOWS_VERSION_8_1, even if the
 * actual Windows environment is newer.
 *
 * [Return value]
 *     Windows version.
 */
extern CONST_FUNCTION int windows_version(void);

/**
 * windows_version_is_at_least:  Return whether the version of Windows on
 * which the program is running is at least the given version.
 *
 * [Parameters]
 *     version: Windows version (a WINDOWS_VERSION_* constant).
 * [Return value]
 *     True if the runtime Windows version is equal to or later than the
 *     given version, false if not.
 */
extern CONST_FUNCTION int windows_version_is_at_least(int version);

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SRC_SYSDEP_WINDOWS_INTERNAL_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
