/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/linux/wrap-io.c: Wrappers for I/O system calls allowing
 * replacement of those calls by override functions.
 */

/* These functions cannot be built with Clang because they make use of the GCC
 * __builtin_apply() set of functions, which are not supported by Clang. */
#ifndef __clang__  // To the end of the file.

#include "src/base.h"
#include "src/sysdep.h"
#include "src/sysdep/linux/internal.h"
#include "src/test/base.h"
#include "src/test/sysdep/linux/wrap-io.h"

#include <dlfcn.h>
#define RTLD_DEFAULT  NULL
#define RTLD_NEXT     ((void *)(intptr_t)-1)

#include <dirent.h>
#include <fcntl.h>
#include <sys/inotify.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utime.h>

/* The trampolines and wrappers here all use __builtin_apply() instead of
 * naming arguments directly, so we need to turn off unused parameter
 * checking. */
#if defined(__clang__)
# pragma clang diagnostic ignored "-Wunused-parameter"
#elif defined(__GNUC__)
# pragma GCC diagnostic ignored "-Wunused-parameter"
#endif

/*************************************************************************/
/*************************************************************************/

#define MAX_ARG_SIZE  (8 * sizeof(void *))
#define DEFINE_WRAPPER(rettype, name, args)                             \
    rettype name args {                                                 \
        if (LIKELY(!override_##name)) {                                 \
            __builtin_return(__builtin_apply((void *)trampoline_##name, \
                                             __builtin_apply_args(),    \
                                             MAX_ARG_SIZE));            \
        } else {                                                        \
            __builtin_return(__builtin_apply((void *)override_##name,   \
                                             __builtin_apply_args(),    \
                                             MAX_ARG_SIZE));            \
        }                                                               \
    }
#define DEFINE_TRAMPOLINE(rettype, name, args)                    \
    rettype trampoline_##name args {                              \
        __typeof__(name) *wrapped_func = dlsym(RTLD_NEXT, #name); \
        ASSERT(wrapped_func != name);                             \
        ASSERT(wrapped_func != NULL);                             \
        __builtin_return(__builtin_apply((void *)wrapped_func,    \
                                         __builtin_apply_args(),  \
                                         MAX_ARG_SIZE));          \
    }
#define DEFINE_FUNCTIONS(rettype, name, args) \
    rettype (*override_##name) args; \
    DEFINE_WRAPPER(rettype, name, args) \
    DEFINE_TRAMPOLINE(rettype, name, args)

DEFINE_FUNCTIONS(int, open, (const char *pathname, int flags, ...))
DEFINE_FUNCTIONS(int, close, (int fd))
DEFINE_FUNCTIONS(ssize_t, read, (int fd, void *buf, size_t count))
DEFINE_FUNCTIONS(ssize_t, write, (int fd, const void *buf, size_t count))
DEFINE_FUNCTIONS(int, utime,
                 (const char *filename, const struct utimbuf *times))
DEFINE_FUNCTIONS(int, fcntl, (int fd, int cmd, ...))
DEFINE_FUNCTIONS(int, ioctl, (int fd, unsigned long request, ...))
DEFINE_FUNCTIONS(ssize_t, readlink,
                 (const char *pathname, char *buf, size_t bufsiz))
DEFINE_FUNCTIONS(int, mkdir, (const char *pathname, mode_t mode))
DEFINE_FUNCTIONS(int, chdir, (const char *path))
DEFINE_FUNCTIONS(DIR *, opendir, (const char *pathname))
DEFINE_FUNCTIONS(struct dirent *, readdir, (DIR *d))
DEFINE_FUNCTIONS(int, closedir, (DIR *d))
DEFINE_FUNCTIONS(int, inotify_init, (void))
DEFINE_FUNCTIONS(int, inotify_add_watch,
                 (int fd, const char *pathname, uint32_t mask))

/*-----------------------------------------------------------------------*/

/* glibc wrapper for read() on a fixed-size buffer.  This might not be
 * declared (such as when compiling at -O0), so we declare it ourselves
 * to avoid compiler warnings. */
extern ssize_t __read_chk(int fd, void *buf, size_t count, size_t buflen);
ssize_t __read_chk(int fd, void *buf, size_t count, size_t buflen)
{
    if (LIKELY(!override_read)) {
        __typeof__(__read_chk) *wrapped_func = dlsym(RTLD_NEXT, "__read_chk");
        ASSERT(wrapped_func != __read_chk);
        ASSERT(wrapped_func != NULL);
        __builtin_return(__builtin_apply((void *)wrapped_func,
                                         __builtin_apply_args(),
                                         MAX_ARG_SIZE));
    } else {
        return read(fd, buf, count);
    }
}

/*************************************************************************/
/*************************************************************************/

#endif  // !__clang__

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
