/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/linux/wrap-x11.c: Wrappers for X11 functions allowing
 * failure injection and call counting.
 */

/* These functions cannot be built with Clang because they make use of the GCC
 * __builtin_apply() set of functions, which are not supported by Clang. */
#ifndef __clang__  // To the end of the file.

#include "src/base.h"
#include "src/sysdep.h"
#include "src/sysdep/linux/internal.h"
#include "src/test/base.h"
#include "src/test/sysdep/linux/wrap-x11.h"

#include <dlfcn.h>
#define RTLD_DEFAULT  NULL
#define RTLD_NEXT     ((void *)(intptr_t)-1)

#include <time.h>
#include <GL/glx.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/extensions/XInput2.h>
#include <X11/extensions/Xinerama.h>
#include <X11/extensions/Xrandr.h>
#include <X11/extensions/xf86vmode.h>

/* The trampolines and wrappers here all use __builtin_apply() instead of
 * naming arguments directly, so we need to turn off unused parameter
 * checking. */
#if defined(__clang__)
# pragma clang diagnostic ignored "-Wunused-parameter"
#elif defined(__GNUC__)
# pragma GCC diagnostic ignored "-Wunused-parameter"
#endif

/*************************************************************************/
/*************************** Data for wrappers ***************************/
/*************************************************************************/

/* Flags for causing wrapped calls to fail. */
uint8_t disable_XCreateBitmapFromData;
uint8_t disable_XCreateColormap;
uint8_t disable_XCreateGC;
uint8_t disable_XCreateIC;
uint8_t disable_XCreatePixmap;
uint8_t disable_XCreatePixmapCursor;
uint8_t disable_XCreateWindow;
uint8_t disable_XGetImage;
uint8_t disable_XLoadQueryFont;
uint8_t disable_XQueryPointer;
uint8_t disable_XF86VidModeQueryExtension;
uint8_t disable_XF86VidModeGetAllModeLines;
uint8_t disable_XF86VidModeGetModeLine;
uint8_t disable_XIQueryVersion;
uint8_t disable_XRRQueryExtension;
uint8_t disable_XRRQueryVersion;
uint8_t disable_XRRGetCrtcInfo;
uint8_t disable_XRRGetOutputInfo;
uint8_t disable_XRRGetPanning;
uint8_t disable_XRRGetScreenResources;
uint8_t disable_XineramaQueryExtension;
uint8_t disable_XineramaIsActive;
uint8_t disable_glXQueryExtension;
uint8_t disable_glXCreateWindow;
uint8_t disable_glXCreateNewContext;
uint8_t disable_glXMakeContextCurrent;

/* Number of calls after which to disable specific wrapped calls. */
uint8_t disable_XCreateGC_after;

/* Flags for generating X11 errors from wrapped calls. */
uint8_t error_XCreateIC;
uint8_t error_XCreateWindow;
uint8_t error_XMoveResizeWindow;

/* Counters indicating the number of times certain functions were called. */
int called_XCreateWindow;
int called_XMoveResizeWindow;
int called_XMoveWindow;
int called_XResetScreenSaver;
int called_XResizeWindow;
int called_XF86VidModeGetAllModeLines;
int called_XF86VidModeGetModeLine;
int called_XRRGetCrtcInfo;
int called_XRRGetOutputInfo;
int called_XRRGetPanning;
int called_XRRGetScreenResources;
int called_XRRSetCrtcConfig;
int called_XineramaIsActive;
int called_XineramaQueryScreens;

/* A copy of the last event sent via XSendEvent() and associated parameters. */
XEvent last_event;
Display *last_event_display;
Window last_event_window;
Bool last_event_propagate;
long last_event_mask;

/* Override return string for Xutf8LookupString() (NULL = no override). */
const char *Xutf8LookupString_override;

/* Versions to report from XIQueryVersion() and XRRQueryVersion(),
 * overriding what the system returns.  Only used if at least one of the
 * relevant {major,minor} pair is nonzero. */
uint8_t xinput_version_major;
uint8_t xinput_version_minor;
uint8_t xrandr_version_major;
uint8_t xrandr_version_minor;

/* Client version reported in the most recent call to XIQueryVersion(). */
int xinput_client_major;
int xinput_client_minor;

/* Flag indicating whether to indicate the presence of a touchscreen in
 * XIQueryDevice(): 1 to indicate that a touchscreen is present, 0 to
 * indicate that no touchscreen is present, or -1 to leave the device list
 * returned by the system unmodified. */
int8_t xinput_simulate_touchscreen = -1;

/*************************************************************************/
/********************** Exported utility functions ***********************/
/*************************************************************************/

void clear_x11_wrapper_variables(void)
{
    disable_XCreateBitmapFromData = 0;
    disable_XCreateColormap = 0;
    disable_XCreateGC = 0;
    disable_XCreateIC = 0;
    disable_XCreatePixmap = 0;
    disable_XCreatePixmapCursor = 0;
    disable_XCreateWindow = 0;
    disable_XGetImage = 0;
    disable_XLoadQueryFont = 0;
    disable_XQueryPointer = 0;
    disable_XF86VidModeQueryExtension = 0;
    disable_XF86VidModeGetAllModeLines = 0;
    disable_XF86VidModeGetModeLine = 0;
    disable_XIQueryVersion = 0;
    disable_XRRQueryExtension = 0;
    disable_XRRQueryVersion = 0;
    disable_XRRGetCrtcInfo = 0;
    disable_XRRGetOutputInfo = 0;
    disable_XRRGetPanning = 0;
    disable_XRRGetScreenResources = 0;
    disable_XineramaQueryExtension = 0;
    disable_XineramaIsActive = 0;
    disable_glXQueryExtension = 0;
    disable_glXCreateWindow = 0;
    disable_glXCreateNewContext = 0;
    disable_glXMakeContextCurrent = 0;

    disable_XCreateGC_after = 0;

    error_XCreateIC = 0;
    error_XCreateWindow = 0;
    error_XMoveResizeWindow = 0;

    called_XCreateWindow = 0;
    called_XMoveResizeWindow = 0;
    called_XMoveWindow = 0;
    called_XResetScreenSaver = 0;
    called_XResizeWindow = 0;
    called_XF86VidModeGetAllModeLines = 0;
    called_XF86VidModeGetModeLine = 0;
    called_XRRGetCrtcInfo = 0;
    called_XRRGetOutputInfo = 0;
    called_XRRGetPanning = 0;
    called_XRRGetScreenResources = 0;
    called_XRRSetCrtcConfig = 0;
    called_XineramaIsActive = 0;
    called_XineramaQueryScreens = 0;

    mem_clear(&last_event, sizeof(last_event));
    last_event_display = NULL;
    last_event_window = None;
    last_event_propagate = False;
    last_event_mask = 0;

    Xutf8LookupString_override = NULL;

    xinput_version_major = 0;
    xinput_version_minor = 0;
    xrandr_version_major = 0;
    xrandr_version_minor = 0;

    xinput_client_major = 0;
    xinput_client_minor = 0;

    xinput_simulate_touchscreen = -1;
}

/*************************************************************************/
/*************************** Wrapper functions ***************************/
/*************************************************************************/

/* Macros which call the wrapped function and return its value from the
 * wrapper.  TRAMPOLINE requires that the function be present;
 * TRAMPOLINE_OPT allows the function to be missing, returning zero in
 * that case. */

#define MAX_ARG_SIZE  (16 * sizeof(void *))

#define TRAMPOLINE(name) {                           \
    void *wrapped_func = dlsym(RTLD_NEXT, #name);    \
    ASSERT(wrapped_func != name);                    \
    ASSERT(wrapped_func != NULL);                    \
    __builtin_return(__builtin_apply(wrapped_func, __builtin_apply_args(), \
                                     MAX_ARG_SIZE)); \
} while (0)

#define TRAMPOLINE_OPT(name)  do {                   \
    void *wrapped_func = dlsym(RTLD_NEXT, #name);    \
    ASSERT(wrapped_func != name);                    \
    if (!wrapped_func) {                             \
        return 0;                                    \
    }                                                \
    __builtin_return(__builtin_apply(wrapped_func, __builtin_apply_args(), \
                                     MAX_ARG_SIZE)); \
} while (0)

/*-------------------------- Core X11 wrappers --------------------------*/

Pixmap XCreateBitmapFromData(Display *display, Drawable d, const char *data,
                             unsigned int width, unsigned int height)
{
    if (disable_XCreateBitmapFromData) {
        return None;
    }
    TRAMPOLINE(XCreateBitmapFromData);
}

Colormap XCreateColormap(Display *display, Window w, Visual *visual, int alloc)
{
    if (disable_XCreateColormap) {
        return None;
    }
    TRAMPOLINE(XCreateColormap);
}

GC XCreateGC(Display *display, Drawable d, unsigned long valuemask,
             XGCValues *values)
{
    if (disable_XCreateGC) {
        return None;
    }
    if (disable_XCreateGC_after > 0) {
        disable_XCreateGC_after--;
        if (disable_XCreateGC_after == 0) {
            disable_XCreateGC = 1;
        }
    }
    TRAMPOLINE(XCreateGC);
}

/* im is actually used, but only indirectly through __builtin_apply_args(),
 * so we need UNUSED to avoid an unused-parameter warning. */
XIC XCreateIC(UNUSED XIM im, ...)
{
    if (disable_XCreateIC) {
        return None;
    }
    if (error_XCreateIC) {
        XGetWindowAttributes(linux_x11_display(), None,
                             (XWindowAttributes[1]){0});
    }
    TRAMPOLINE(XCreateIC);
}

Pixmap XCreatePixmap(Display *display, Drawable d, unsigned int width,
                     unsigned int height, unsigned int depth)
{
    if (disable_XCreatePixmap) {
        return None;
    }
    TRAMPOLINE(XCreatePixmap);
}

Cursor XCreatePixmapCursor(Display *display, Pixmap source, Pixmap mask,
                           XColor *foreground_color, XColor *background_color,
                           unsigned int x, unsigned int y)
{
    if (disable_XCreatePixmapCursor) {
        return None;
    }
    TRAMPOLINE(XCreatePixmapCursor);
}

Window XCreateWindow(Display *display, Window parent, int x, int y,
                     unsigned int width, unsigned int height,
                     unsigned int border_width, int depth, unsigned int class,
                     Visual *visual, unsigned long valuemask,
                     XSetWindowAttributes *attributes)
{
    called_XCreateWindow++;
    if (disable_XCreateWindow) {
        return None;
    }
    if (error_XCreateWindow) {
        XGetWindowAttributes(linux_x11_display(), None,
                             (XWindowAttributes[1]){0});
    }
    TRAMPOLINE(XCreateWindow);
}

XImage *XGetImage(Display *display, Drawable d, int x, int y,
                  unsigned int width, unsigned int height,
                  unsigned long plane_mask, int format)
{
    if (disable_XGetImage) {
        return None;
    }
    TRAMPOLINE(XGetImage);
}

XFontStruct *XLoadQueryFont(Display *display, const char *name)
{
    if (disable_XLoadQueryFont) {
        return NULL;
    }
    TRAMPOLINE(XLoadQueryFont);
}

int XMoveResizeWindow(Display *display, Window w, int x, int y,
                      unsigned int width, unsigned int height)
{
    called_XMoveResizeWindow++;
    if (error_XMoveResizeWindow) {
        XGetWindowAttributes(linux_x11_display(), None,
                             (XWindowAttributes[1]){0});
        /* The return value is not documented, but libX11 always returns 1. */
        return 1;
    }
    TRAMPOLINE(XMoveResizeWindow);
}

int XMoveWindow(Display *display, Window w, int x, int y)
{
    called_XMoveWindow++;
    TRAMPOLINE(XMoveWindow);
}

Bool XQueryPointer(Display *display, Window w, Window *root_return,
                   Window *child_return, int *root_x_return,
                   int *root_y_return, int *win_x_return, int *win_y_return,
                   unsigned int *mask_return)
{
    if (disable_XQueryPointer) {
        return False;
    }
    TRAMPOLINE(XQueryPointer);
}

int XResetScreenSaver(Display *display)
{
    called_XResetScreenSaver++;
    TRAMPOLINE(XResetScreenSaver);
}

int XResizeWindow(Display *display, Window w, unsigned int width,
                  unsigned int height)
{
    called_XResizeWindow++;
    TRAMPOLINE(XResizeWindow);
}

Bool XSendEvent(Display *display, Window w, Bool propagate, long event_mask,
                XEvent *event_send)
{
    last_event_display = display;
    last_event_window = w;
    last_event_propagate = propagate;
    last_event_mask = event_mask;
    last_event = *event_send;
    TRAMPOLINE(XSendEvent);
}

int Xutf8LookupString(XIC ic, XKeyPressedEvent *event, char *buffer_return,
                      int bytes_buffer, KeySym *keysym_return,
                      Status *status_return)
{
    if (Xutf8LookupString_override) {
        if ((int)strlen(Xutf8LookupString_override) > bytes_buffer) {
            if (status_return) {
                *status_return = XBufferOverflow;
            }
        } else {
            memcpy(buffer_return, Xutf8LookupString_override,
                   strlen(Xutf8LookupString_override));
            if (status_return) {
                *status_return = XLookupChars;
            }
        }
        return strlen(Xutf8LookupString_override);
    }
    TRAMPOLINE(Xutf8LookupString);
}

/*------------------------ XF86VidMode wrappers -------------------------*/

Bool XF86VidModeQueryExtension(Display *dpy, int *event_base_return,
                               int *error_base_return)
{
    if (disable_XF86VidModeQueryExtension) {
        return False;
    }
    TRAMPOLINE_OPT(XF86VidModeQueryExtension);
}

Bool XF86VidModeGetAllModeLines(Display *dpy, int screen,
                                int *modecount_return,
                                XF86VidModeModeInfo ***modelinesPtr_return)
{
    called_XF86VidModeGetAllModeLines++;
    if (disable_XF86VidModeGetAllModeLines) {
        return False;
    }
    TRAMPOLINE(XF86VidModeGetAllModeLines);
}

Bool XF86VidModeGetModeLine(Display *dpy, int screen, int *dotclock_return,
                            XF86VidModeModeLine *modeline_return)
{
    called_XF86VidModeGetModeLine++;
    if (disable_XF86VidModeGetModeLine) {
        return False;
    }
    TRAMPOLINE(XF86VidModeGetModeLine);
}

/*-------------------------- XInput2 wrappers ---------------------------*/

Status XIQueryVersion(Display *display, int *major_version_inout,
                      int *minor_version_inout)
{
    if (disable_XIQueryVersion) {
        return BadRequest;
    }
    __typeof__(XIQueryVersion) *p_XIQueryVersion =
        dlsym(RTLD_NEXT, "XIQueryVersion");
    ASSERT(p_XIQueryVersion != XIQueryVersion);
    if (!p_XIQueryVersion) {
        return BadRequest;
    }
    xinput_client_major = *major_version_inout;
    xinput_client_minor = *minor_version_inout;
    Status result = (*p_XIQueryVersion)(display, major_version_inout,
                                        minor_version_inout);
    if (result != Success) {
        return result;
    }
    if (xinput_version_major || xinput_version_minor) {
        *major_version_inout = xinput_version_major;
        *minor_version_inout = xinput_version_minor;
    }
    return Success;
}

XIDeviceInfo *XIQueryDevice(Display *display, int deviceid,
                            int *ndevices_return)
{
    __typeof__(XIQueryDevice) *p_XIQueryDevice =
        dlsym(RTLD_NEXT, "XIQueryDevice");
    ASSERT(p_XIQueryDevice);
    ASSERT(p_XIQueryDevice != XIQueryDevice);
    XIDeviceInfo *devices =
        (*p_XIQueryDevice)(display, deviceid, ndevices_return);
    if (devices && xinput_simulate_touchscreen >= 0) {
        int found = 0;
        for (int i = 0; i < *ndevices_return; i++) {
            if (devices[i].use == XIMasterPointer) {
                found = 1;
                if (xinput_simulate_touchscreen) {
                    int j;
                    for (j = 0; j < devices[i].num_classes; j++) {
                        XITouchClassInfo *class =
                            (XITouchClassInfo *)devices[i].classes[j];
                        if (class->type == XITouchClass
                         && class->mode == XIDirectTouch) {
                            break;
                        }
                    }
                    if (j == devices[i].num_classes) {
                        /* No touchscreen, so fake one.  We don't check
                         * any classes other than XITouchClass, and
                         * XFreeDeviceList() doesn't care if we modify this
                         * data, so just arbitrarily overwrite the last
                         * class. */
                        ASSERT(j > 0);
                        XITouchClassInfo *class =
                            (XITouchClassInfo *)devices[i].classes[j-1];
                        class->type = XITouchClass;
                        class->mode = XIDirectTouch;
                        class->num_touches = 5;  // Arbitrary.
                    }
                } else {
                    for (int j = 0; j < devices[i].num_classes; j++) {
                        XITouchClassInfo *class =
                            (XITouchClassInfo *)devices[i].classes[j];
                        if (class->type == XITouchClass
                         && class->mode == XIDirectTouch) {
                            /* Null out the entry so this isn't detected as
                             * a touchscreen. */
                            devices[i].classes[j]->type = -1;
                        }
                    }
                }
            }
        }
        ASSERT(found);
    }
    return devices;
}

/*--------------------------- XRandR wrappers ---------------------------*/

Bool XRRQueryExtension(Display *dpy, int *event_base_return,
                       int *error_base_return)
{
    if (disable_XRRQueryExtension) {
        return False;
    }
    TRAMPOLINE_OPT(XRRQueryExtension);
}

Status XRRQueryVersion(Display *dpy, int *major_version_return,
                       int *minor_version_return)
{
    if (disable_XRRQueryVersion) {
        /* The return value is prototyped as "Status", but it's really Bool. */
        return False;
    }
    __typeof__(XRRQueryVersion) *p_XRRQueryVersion =
        dlsym(RTLD_NEXT, "XRRQueryVersion");
    ASSERT(p_XRRQueryVersion != XRRQueryVersion);
    if (!p_XRRQueryVersion) {
        return False;
    }
    if (!(*p_XRRQueryVersion)(dpy, major_version_return,
                              minor_version_return)) {
        return False;
    }
    if (xrandr_version_major || xrandr_version_minor) {
        *major_version_return = xrandr_version_major;
        *minor_version_return = xrandr_version_minor;
    }
    return True;
}

XRRCrtcInfo *XRRGetCrtcInfo(Display *dpy, XRRScreenResources *resources,
                            RRCrtc crtc)
{
    called_XRRGetCrtcInfo++;
    if (disable_XRRGetCrtcInfo) {
        return NULL;
    }
    TRAMPOLINE(XRRGetCrtcInfo);
}

XRROutputInfo *XRRGetOutputInfo(Display *dpy, XRRScreenResources *resources,
                                RROutput output)
{
    called_XRRGetOutputInfo++;
    if (disable_XRRGetOutputInfo) {
        return NULL;
    }
    TRAMPOLINE(XRRGetOutputInfo);
}

XRRPanning *XRRGetPanning(Display *dpy, XRRScreenResources *resources,
                          RRCrtc crtc)
{
    called_XRRGetPanning++;
    if (disable_XRRGetPanning) {
        return NULL;
    }
    TRAMPOLINE(XRRGetPanning);
}

XRRScreenResources *XRRGetScreenResources(Display *dpy, Window window)
{
    called_XRRGetScreenResources++;
    if (disable_XRRGetScreenResources) {
        return NULL;
    }
    TRAMPOLINE(XRRGetScreenResources);
}

Status XRRSetCrtcConfig(Display *dpy, XRRScreenResources *resources,
                        RRCrtc crtc, Time timestamp, int x, int y, RRMode mode,
                        Rotation rotation, RROutput *outputs, int noutputs)
{
    called_XRRSetCrtcConfig++;
    TRAMPOLINE(XRRSetCrtcConfig);
}

/*-------------------------- Xinerama wrappers --------------------------*/

Bool XineramaQueryExtension(Display *dpy, int *event_base_return,
                               int *error_base_return)
{
    if (disable_XineramaQueryExtension) {
        return False;
    }
    TRAMPOLINE_OPT(XineramaQueryExtension);
}

Bool XineramaIsActive(Display *dpy)
{
    called_XineramaIsActive++;
    if (disable_XineramaIsActive) {
        return False;
    }
    TRAMPOLINE(XineramaIsActive);
}

XineramaScreenInfo *XineramaQueryScreens(Display *dpy, int *number_return)
{
    called_XineramaQueryScreens++;
    TRAMPOLINE(XineramaQueryScreens);
}

/*---------------------------- glX wrappers -----------------------------*/

Bool glXQueryExtension(Display *dpy, int *event_base_return,
                       int *error_base_return)
{
    if (disable_glXQueryExtension) {
        return False;
    }
    TRAMPOLINE_OPT(glXQueryExtension);
}

GLXWindow glXCreateWindow(Display *dpy, GLXFBConfig config, Window win,
                          const int *attribList)
{
    if (disable_glXCreateWindow) {
        return None;
    }
    TRAMPOLINE(glXCreateWindow);
}

GLXContext glXCreateNewContext(Display *dpy, GLXFBConfig config,
                               int renderType, GLXContext shareList,
                               Bool direct)
{
    if (disable_glXCreateNewContext) {
        return None;
    }
    TRAMPOLINE(glXCreateNewContext);
}

Bool glXMakeContextCurrent(Display *dpy, GLXDrawable draw,
                           GLXDrawable read, GLXContext ctx)
{
    if (disable_glXMakeContextCurrent) {
        return False;
    }
    TRAMPOLINE(glXMakeContextCurrent);
}

/*************************************************************************/
/*************************************************************************/

#endif  // !__clang__

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
